# coding=utf-8
"""QGIS related build tools."""
from fabric.contrib.files import exists
from fabric.api import run, cd, env, task, sudo
import fabtools
from .common import add_ubuntugis_ppa
from .common import setup_env
from .git import update_git_checkout
from .system import setup_ccache
from .gdal import build_gdal
from .postgres import create_postgis_1_5_db


@task
def clone_qgis(branch='master', delete_local_branches=False):
    """Clone or update QGIS from git.

    :param branch: Name of the branch to build from. Defaults to 'master'
    :type branch: str

    :param delete_local_branches: Whether existing local branches should be
        pruned away from git.
    :type delete_local_branches: bool
    """
    setup_env()
    fabtools.require.deb.package('git')
    # Add this to the users git config so that we don't get repeated
    # authentication requests when using ssl
    #run('git config --global credential.helper \'cache --timeout=3600\'')
    #run('git config --global push.default simple')

    code_base = '%s/cpp' % env.fg.workspace
    code_path = '%s/QGIS' % code_base

    update_git_checkout(
        code_base,
        env.fg.qgis_git_url,
        'QGIS',
        branch)
    if exists(code_path):
        with cd(code_path):
            if delete_local_branches:
                run('git branch | grep -v \* | xargs git branch -D')


def compile_qgis(build_path, build_prefix, gdal_from_source=False):
    """Compile QGIS including installation of built tools and dependencies.


    :param build_path: Path to the cmake build dir that should be used. Path
        must point to under the QGIS git checkout dir. e.g.
        QGIS/build-fabgis.
    :type build_path: str

    :param build_prefix: Path where the QGIS binaries should be installed to.
        e.g. /usr/local/qgis-master
    :type build_prefix: str

    :param gdal_from_source: Whether gdal should be built from source.
        Default False.
    :type gdal_from_source: bool
    """

    fabtools.require.deb.package('cmake-curses-gui')
    fabtools.require.deb.package('grass-dev')
    fabtools.require.deb.package('grass')
    fabtools.require.deb.package('git')
    fabtools.require.deb.package('python-gdal')
    fabtools.require.deb.package('libfcgi-dev')
    # Ensure we always have a clean build dir
    if exists(build_path):
        run('rm -rf %s' % build_path)
    fabtools.require.directory(build_path)
    with cd(build_path):
        fabtools.require.directory(
            build_prefix,
            use_sudo=True,
            owner=env.fg.user)
        os_version = run('cat /etc/issue.net')
        os_version = float(os_version.split(' ')[1])

        if os_version > 13:
            extra = '-DPYTHON_LIBRARY=/usr/lib/x86_64-linux-gnu/libpython2.7.so'
        else:
            extra = ''

        if gdal_from_source:
            build_gdal()  # see that task for ecw and mrsid support
            extra += '-DGDAL_CONFIG=/usr/local/bin/gdal-config '

        cmake = ('cmake .. '
                 '-DCMAKE_INSTALL_PREFIX=%s '
                 '-DCMAKE_CXX_COMPILER:FILEPATH=/usr/local/bin/g++ '
                 '-DQT_QMAKE_EXECUTABLE=/usr/bin/qmake-qt4 '
                 '-DWITH_MAPSERVER=ON '
                 '-DWITH_INTERNAL_SPATIALITE=ON '
                 '-DWITH_GRASS=OFF '
                 '%s'
                 % (build_prefix, extra))
        run('cmake .. %s' % cmake)
        processor_count = run('cat /proc/cpuinfo | grep processor | wc -l')
        run('time make -j %s install' % processor_count)


@task
def install_qgis1_8(gdal_from_source=False):
    """
    Install QGIS 1.8 under /usr/local/qgis-1.8.

    :param gdal_from_source: Whether gdal should be built from source.
        Default False.
    :type gdal_from_source: bool


    """
    setup_env()
    add_ubuntugis_ppa()
    setup_ccache()
    sudo('apt-get build-dep -y qgis')
    clone_qgis(branch='release-1_8')
    workspace = '%s/cpp' % env.fg.workspace
    code_path = '%s/QGIS' % workspace
    build_path = '%s/build-qgis18-fabgis' % code_path
    build_prefix = '/usr/local/qgis-1.8'
    compile_qgis(build_path, build_prefix, gdal_from_source)


@task
def install_qgis2(gdal_from_source=False):
    """Install QGIS 2 under /usr/local/qgis-2.0.

    :param gdal_from_source: Whether gdal should be built from source.
        Default False.
    :type gdal_from_source: bool

    TODO: create one function from this and the 1.8 function above for DRY.

    """
    setup_env()
    setup_ccache()
    add_ubuntugis_ppa()
    sudo('apt-get build-dep -y qgis')

    #fabtools.require.deb.package('python-pyspatialite')
    fabtools.require.deb.package('python-psycopg2')
    fabtools.require.deb.package('python-qscintilla2')
    fabtools.require.deb.package('libqscintilla2-dev')
    fabtools.require.deb.package('libspatialindex-dev')

    clone_qgis(branch='release-2_0')
    workspace = '%s/cpp' % env.fg.workspace
    code_path = '%s/QGIS' % workspace
    build_path = '%s/build-qgis2-fabgis' % code_path
    build_prefix = '/usr/local/qgis-2.0'
    compile_qgis(build_path, build_prefix, gdal_from_source)


@task
def install_qgis_master(gdal_from_source=False):
    """Install QGIS master under /usr/local/qgis-master.

    :param gdal_from_source: Whether gdal should be built from source.
        Default False.
    :type gdal_from_source: bool

    TODO: create one function from this and the 1.8 function above for DRY.

    """
    setup_env()
    setup_ccache()
    add_ubuntugis_ppa()
    sudo('apt-get build-dep -y qgis')

    #fabtools.require.deb.package('python-pyspatialite')
    fabtools.require.deb.package('python-psycopg2')
    fabtools.require.deb.package('python-qscintilla2')
    fabtools.require.deb.package('libqscintilla2-dev')
    fabtools.require.deb.package('libspatialindex-dev')

    clone_qgis(branch='master')
    workspace = '%s/cpp' % env.fg.workspace
    code_path = '%s/QGIS' % workspace
    build_path = '%s/build-master-fabgis' % code_path
    build_prefix = '/usr/local/qgis-master'
    compile_qgis(build_path, build_prefix, gdal_from_source)


@task
def setup_qgis2_and_postgis():
    """
    Install qgis2 and postgis 1.5 as a fully working setup
    """
    create_postgis_1_5_db('gis', env.user)
    install_qgis2()



# coding=utf-8
"""Helpers dropbox so that you can easily move gis data to the server."""

import os

from fabric.contrib.files import exists
from fabtools import require

from fabric.api import run, cd, env, task, sudo, put
from .common import setup_env
from .utilities import replace_tokens
from .utilities import append_if_not_present


@task
def setup_dropbox():
    """Setup a headless dropbox installation."""
    setup_env()
    dropbox_url = 'https://www.dropbox.com/download?plat=lnx.x86_64'
    dropbox_cli_url = 'https://www.dropbox.com/download?dl=packages/dropbox.py'

    with cd('~'):
        require.directory('bin')
        with cd('bin'):
            if not exists('dropbox.py'):
                run('wget -O dropbox.py "%s"' % dropbox_cli_url)
                run('chmod +x dropbox.py')
        if not exists('.dropbox-dist'):
            command = 'wget -O - "%s" | tar xzf -' % dropbox_url
            run(command)
        # After you have pasted the link and confirmed the machine account
        # you need to press Ctrl-c which unfortunately also terminates the
        # fabric session. capture=False is an attempt to remedy that.
        run('~/.dropbox-dist/dropboxd')


@task
def setup_dropbox_daemon():
    """Set up dropbox to run on init.d.

    Unfortunately we can't run this as part of initial setup due to the
    control-c issue that prevents us from halting the initial dropboxd
    command needed to set up the dropbox synced account.

    So for example you need to do::

        fab vagrant setup_dropbox

    Then once your account is linked, press Ctrl-c - which will also
    terminate the above fabric job. Then run::

        fab vagrant setup_dropbox_daemon

    """
    setup_env()
    with cd('/etc/init.d/'):
        if not exists('dropbox'):
            local_dir = os.path.dirname(__file__)
            local_file = os.path.abspath(os.path.join(
                local_dir,
                '..',
                '../fabgis_resources',
                'server_config',
                'dropbox',
                'dropbox.templ'))
            put(local_file,
                '/etc/init.d/dropbox',
                use_sudo=True)

        my_tokens = {'USER': env.fg.user, }
        replace_tokens('/etc/init.d/dropbox', my_tokens)

    sudo('chmod +x /etc/init.d/dropbox')
    sudo('update-rc.d dropbox defaults')
    start_dropbox()
    dropbox_status()


@task
def start_dropbox():
    """Start the dropbox service."""
    sudo('service dropbox start')


@task
def stop_dropbox():
    """Stop the dropbox service."""
    sudo('service dropbox stop')


@task
def dropbox_status():
    """Check the status of dropbox."""
    run('~/bin/dropbox.py status')



# coding=utf-8
"""Tools for deployment of tilestream.

.. note:: Usually you should run tilemill on your local host and prepare your
    mbtiles, then use tilestream (https://github.com/mapbox/tilestream) to
    host the resulting mbtiles output.

.. seealso:: :file:`tilemill.py`

"""
import os

from fabric.api import fastprint, task, sudo, run, cd, env, put
from fabtools.files import exists
from fabtools import require
from .common import setup_env
from .utilities import replace_tokens
from .system import get_ip_address


@task
def setup_tilestream():
    """Set up tile stream - see https://github.com/mapbox/tilestream.

    This one deserves a little explanation:

    Tilestream is a nodejs application. Node seems to be pretty unparticular
    about maintaining api compatibility between releases so if you grab one
    from e.g. apt, chances are it won't work with tilestream.

    To address this, we use the nodeenv virtualisation environment (somewhat
    equivalent to using python virtualenv) to ensure that we have the
    expected version of tilestream. e.g.::

        nodeenv env --node=0.8.15
    """
    setup_env()
    require.deb.package('curl')
    require.deb.package('build-essential')
    require.deb.package('libssl-dev')
    require.deb.package('libsqlite3-0')
    require.deb.package('libsqlite3-dev')
    require.deb.package('git-core')
    require.deb.package('nodejs nodejs-dev npm')
    require.deb.package('python-pip')

    sudo('pip install nodeenv')

    dev_dir = '/home/%s/dev/javascript' % env.fg.user
    fastprint('making directory %s' % dev_dir)
    require.directory(dev_dir)

    tile_stream_dir = os.path.join(dev_dir, 'tilestream')
    fastprint('checkout out tilestream to %s' % tile_stream_dir)

    if not exists(tile_stream_dir):
        with cd(dev_dir):
            run('git clone http://github.com/mapbox/tilestream.git')

    with cd(tile_stream_dir):
        if not exists(os.path.join(tile_stream_dir, 'env')):
            run('nodeenv env --node=0.8.15')
        # If doing this interactively from a shell you would first do:
        # . env/bin/activate
        # npm install
        # From our scripted environment (where we cant activate the venv):
        run('env/bin/npm install')


@task
def setup_tilestream_daemon(
        tile_dir=None, ui_port=8888, tiles_port=8888, tile_host=None):
    """Start the tilestream service - ensure it is installed first.

    .. warning:: This task does not work yet.

    :param tile_dir: Optional directory on the remote tile_host that holds one
        or more mbtiles files to be published. If ommitted the default of
        `~/Documents/MapBox/tiles` will be used by tilestream. Be sure to
        escape and paths given here. e.g.::

            fab vagrant setup_tilestream_daemon:tile_dir='\/vagrant'

    :type tile_dir: str

    :param ui_port: Port on which the tilestream ui should be available.
    :type ui_port: int

    :param tiles_port: Port on which tilestream tile service should be
        available. You may want to use a different port here and then run a
        CDN such as cloudflare in front of the tile service.
    :type tiles_port: int

    :param tile_host: Host name under which the service will run. Must match the
        hostname to which requests are made (even if you run it behind mod
        proxy). If none it will be set to the IP address of the tile_host.
    :type tile_host: str
    """
    setup_env()
    params = _build_parameters(tile_host, tile_dir, tiles_port, ui_port)
    template_file = 'tilestream.templ'
    template_path = os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        os.path.pardir,
        '../fabgis_resources',
        'server_config',
        'tilestream',
        '%s' % template_file
    ))
    remote_path = '/tmp/%s' % template_file
    fastprint(template_path, remote_path)
    put(template_path, remote_path)
    binary = 'index.js'
    if not tile_host:
        tile_host = get_ip_address()
    tokens = {
        'BINARY': binary,
        'OPTIONS': params,
        'USER': env.fg.user,
        'TILES_DIR': tile_dir,
        'TILES_PORT': tiles_port,
        'UI_PORT': ui_port,
        'HOST': tile_host}
    replace_tokens(remote_path, tokens)
    sudo('mv /tmp/tilestream /etc/init.d/')
    sudo('chmod 755 /etc/init.d/tilestream')
    sudo('/etc/init.d/tilestream start')
    sudo('sysv -rc-conf --level 2345 tilestream on')


def _build_parameters(host, tile_dir, tiles_port, ui_port):
    """Helper to build the parameters needed for starting tilestream.

    :param tile_dir: Optional directory on the remote host that holds one or
        more mbtiles files to be published. If ommitted the default of
        `~/Documents/MapBox/tiles` will be used by tilestream.
    :type tile_dir: str

    :param ui_port: Port on which the tilestream ui should be available.
    :type ui_port: int

    :param tiles_port: Port on which tilestream tile service should be
        available. You may want to use a different port here and then run a
        CDN such as cloudflare in front of the tile service.
    :type tiles_port: int

    :param host: Host name under which the service will run. Must match the
        hostname to which requests are made (even if you run it behind mod
        proxy).
    :type host: str

    :returns: A string containing parameters for starting tilestream.
    :rtype: str
    """
    params = ''
    if tile_dir is not None:
        params += ' --tiles=%s' % tile_dir
    params += ' --uiPort=%i' % ui_port
    params += ' --tilePort=%i' % tiles_port
    if host is not None:
        params += ' --host=%s' % host
    return params


@task
def start_tilestream(tile_dir=None, ui_port=8888, tiles_port=8888, host=None):
    """Start the tilestream service - ensure it is installed first.

    Use this task when you want to start tilestream manually for testing.

    :param tile_dir: Optional directory on the remote host that holds one or
        more mbtiles files to be published. If ommitted the default of
        `~/Documents/MapBox/tiles` will be used by tilestream.
    :type tile_dir: str

    :param ui_port: Port on which the tilestream ui should be available.
    :type ui_port: int

    :param tiles_port: Port on which tilestream tile service should be
        available. You may want to use a different port here and then run a
        CDN such as cloudflare in front of the tile service.
    :type tiles_port: int

    :param host: Host name under which the service will run. Must match the
        hostname to which requests are made (even if you run it behind mod
        proxy).
    :type host: str

    Example invocation of tilestream with all the bells and whistles::

        PATH=env/bin:$PATH ./index.js \
            --host = 41.74.158.13 \
            --accesslog = /tmp/tilestream.log \
            --uiPort=8888 \
            --tilePort = 8889
    """
    setup_env()
    dev_dir = '/home/%s/dev/javascript' % env.fg.user
    tile_stream_dir = os.path.join(dev_dir, 'tilestream')
    with cd(tile_stream_dir):
        params = _build_parameters(host, tile_dir, tiles_port, ui_port)
        run('PATH=env/bin:$PATH ./index.js %s' % params)

    fastprint(
        'You may need to port forward to port 8888 or set up your '
        'vagrant instance to do so....')



# coding=utf-8
"""
Django related tasks
--------------------
"""

import os
from fabric.api import cd, task, sudo, fastprint, run
from fabric.colors import green
from fabric.contrib.files import sed, upload_template
from fabric.contrib.files import exists
from fabtools import require, fabtools
from .common import setup_env
from . import virtualenv


@task
def set_media_permissions(code_path, wsgi_user='wsgi', media_dir=None):
    """Set the django media dir so apache can write to it.

    :param code_path: Path to top level deploy dir.
    :type code_path: str
    :param media_dir: Optional dir underneath code_path if media does not live in
        ``<code_path>/django_project/media``. No trailing slash.
    :type media_dir: str
    :param wsgi_user: User that should receive write permissions to the folder.
        Defaults to 'wsgi'.
    :type wsgi_user: str
    """
    if not media_dir:
        media_path = '%s/django_project/media' % code_path
    else:
        media_path = '%s/%s' % (code_path, media_dir)
    if not exists(media_path):
        sudo('mkdir %s' % media_path)
    sudo('chgrp -R %s %s' % (wsgi_user, media_path))


@task
def setup_apache(
        site_name,
        code_path,
        domain,
        template_dir=None,
        media_dir=None,
        wsgi_user='wsgi',
        **kwargs):
    """Set up the apache server for this site.

    :param site_name: Name of the site e.g. changelogger. Should be a single
        word with only alpha characters in it.
    :type site_name: str

    :param code_path: Directory where the code lives. Will be used to set
        media etc permissions.
    :type code_path: str

    :param domain: Domain name. If none will be set to hostname.
    :type domain: str

    :param template_dir: Directory where the template files live. If none
        will default to ``resources/server_config/apache``. Must be a
        relative path to the fabfile you are running.
    :type template_dir: str
    
    :param media_dir: Optional dir under code_path if media does not live in 
        ``<code_path>/django_project/media``. No trailing slash.
    :type media_dir: str

    :param wsgi_user: Name of user wsgi process should run as. The user will
        be created as needed.
    :type wsgi_user: str

    :param kwargs: Any extra keyword arguments that should be appended to the
        token list that will be used when rendering the apache config template.
        Use this to pass in sensitive data such as passwords.
    :type kwargs: dict

    :returns: Path to the apache conf file.
    :rtype: str
    """
    setup_env()
    # Ensure we have a mailserver setup for our domain
    # Note that you may have problems if you intend to run more than one
    # site from the same server
    require.postfix.server(site_name)
    require.deb.package('libapache2-mod-wsgi')

    # Find out if the wsgi user exists and create it if needed e.g.
    require.user(
        wsgi_user,
        create_group=wsgi_user,
        system=True,
        comment='System user for running the wsgi process under')

    # Clone and replace tokens in apache conf
    if template_dir is None:
        template_dir = 'resources/server_config/apache/'
    filename = '%s.apache.conf.templ' % site_name
    template_path = os.path.join(template_dir, filename)
    fastprint(green('Using %s for template' % template_path))

    context = {
        'escaped_server_name': domain.replace('.', '\.'),
        'server_name': domain,
        'site_user': wsgi_user,
        'code_path': code_path,
        'site_name': site_name}
    context.update(kwargs)  # merge in any params passed in to this function
    destination = '/etc/apache2/sites-available/%s.apache.conf' % site_name
    fastprint(context)

    upload_template(
        template_path,
        destination,
        context=context,
        use_sudo=True)

    set_media_permissions(code_path, wsgi_user, media_dir=media_dir)

    sudo('a2ensite %s.apache.conf' % site_name)
    sudo('a2dissite default')
    sudo('a2enmod rewrite')
    # Check if apache configs are ok - script will abort if not ok
    sudo('/usr/sbin/apache2ctl configtest')
    require.service.restarted('apache2')
    return destination


@task
def build_pil(code_path):
    """Build pil with proper support for jpeg etc.

    :param code_path: Directory where the code lives.
    :type code_path: str

    .. note:: Any existing PIL will be uninstalled.

    .. versionchanged:: 0.16.0 - delegate to virtualenv module

    .. deprecated:: 0.16.0
       Use :func:`virtualenv.build_pil` rather.
    """
    virtualenv.build_pil(code_path)


@task
def setup_celery(project_name, user, password, code_path):
    """
    Installs and starts django-celery with rabbitmq server

    Copy celery config files to correct locations and start rabbitmq and
    celery servers
    :param password: Password for rabbitmq server
    :param user: User for rabbitmq server
    :param project_name: Name for rabbitmq server vhost
    :param code_path: Directory where the code lives.
    :type code_path: str

    Note: In a production environment, you will want to daemonize the celery
    worker:

    http://docs.celeryproject.org/en/latest/tutorials/daemonizing.html#
    daemonizing
    """
    fabtools.require.deb.package('rabbitmq-server')
    sudo('rabbitmqctl add_user %s %s' % (user, password))
    sudo('rabbitmqctl add_vhost %s' % project_name)
    sudo('rabbitmqctl set_permissions -p %s %s ".*" ".*" ".*"' % (
        project_name, user))
    sudo('rabbitmqctl delete_user guest')
    venv = os.path.join(code_path, 'venv')
    with cd(venv):
        run('bin/pip install django-celery')
    run('celery -A tasks worker --loglevel=info')
    fabtools.require.service.restarted('celeryd')
    fabtools.require.service.restarted('rabbitmq-server')



"""Git related utilities."""
import os
from fabric.api import fastprint, run, cd, task, env
from fabric.colors import red, cyan, green
from fabric.contrib.files import exists
import fabtools

from .common import setup_env


@task
def update_git_checkout(code_path, url, repo_alias, branch='master', tag=None):
    """Make sure there is a read only git checkout.

    :param code_path: Path to where the repo should be.
    :type code_path: str

    :param url: Complete http url for cloning/checking out the repo.
    :type url: str

    :param repo_alias: Alias name under which the repo should be checked out.
    :type repo_alias: str

    :param branch: The name of the branch to build from. Defaults to 'master'.
    :type branch: str

    :param tag: Tag to checkout. If None this is ignored. If not none will be
        used in preference to branch.
    :type tag: None, str

    To run e.g.::

        fab -H 188.40.123.80:8697 remote update_git_checkout

    """
    fastprint(cyan('Updating git checkout.\n'))
    fastprint(cyan('code_path: %s\n' % code_path))
    fastprint(cyan('url: %s\n' % url))
    fastprint(cyan('repo_alias: %s\n' % repo_alias))
    fastprint(cyan('branch: %s\n' % branch))
    fastprint(cyan('tag: %s\n' % tag))
    setup_env()
    repo_path = os.path.join(code_path, repo_alias)

    fabtools.require.deb.package('git')
    if not exists(repo_path):
        fastprint(green('Repository does not exist, creating.\n'))
        fabtools.require.directory(code_path, use_sudo=True, owner=env.user)
        with cd(code_path):
            run('git clone %s %s' % (url, repo_alias))
    else:
        fastprint(green('Repo checkout does exist, updating.'))
        with cd(repo_path):
            # Get any updates first
            run('git fetch')
            if tag is not None:
                tags = run('git tag --list')
                if tag in tags:
                    fastprint(green('Checking out tag.'))
                    run('git checkout %s' % tag)
                else:
                    fastprint(red('Checking out tag failed - unknown tag.'))
                    raise Exception('Git checkout failed for tag')

            elif branch != 'master':
                fastprint(green('Checking out branch.'))
                branches = run('git branch')
                if branch not in branches:
                    print branches
                    run('git branch --track %s origin/%s' % (branch, branch))
                if ('* %s' % branch) not in branches:
                    run('git checkout %s' % branch)
            else:
                fastprint(green('Checking out master.'))
                run('git checkout master')
            run('git pull')
    fastprint(green('Checking out tag completed.'))


@task
def remove_local_branches(code_path):
    """Remove any local branches you may have in your repo - use with caution!

    Any local changes will be dropped. Master will be checked out and all
    other local branches removed.

    :param code_path: Path to where the repo should be.
    :type code_path: str

    .. versionadded:: 0.16.0
    """
    fastprint(red('Removing all local branches from repo! %s\n' % code_path))
    with cd(code_path):
        # Get rid of any local changes
        run('git reset --hard')
        # Get back onto master branch
        run('git checkout master')
        # Delete all local branches
        run('git branch | grep -v \* | xargs git branch -D')



# coding=utf-8
"""Tools for configuration and set up of docker.

.. seealso:: http://www.docker.io/
     and http://docs.docker.io/en/latest/installation/
     ubuntulinux/#ubuntu-precise-12-04-lts-64-bit

As the second article mentions, for ubuntu 12.04 the kernel needs to be
upgraded to 3.8 - which the setup task will do.


Example usage:

In your fabfile.py you might do the initial setup of docker (assuming it isn't
already installed on your system), create a base image with a root password
of your choosing and then bring up a container like this::

    setup_docker()

Once it is set up, you might create a new container for your project::

    setup_docker_image()  # called by setup docker case of first install
    create_docker_container()
    id = current_docker_container()
    ssh_port = get_docker_port_mappings(id)[22]

A control file called `fabgis.container.id` will be written to the current
working directory - it will contain the id of the created container.

After this you can treat the docker container as any other host and run
fabric tasks against it e.g.::

    fab -H root@localhost:<ssh_port> <command>

"""

from fabric.api import run, sudo, task, fastprint, env, prompt, reboot, abort
from fabric.colors import red, green, blue, yellow
from fabric.contrib.files import exists, contains, sed
from fabtools.deb import update_index as apt_get_update, is_installed
from fabtools.require.deb import ppa as require_ppa
from fabtools.require.deb import package as require_package
from fabtools.require.deb import packages as require_packages


@task
def docker():
    """Convenience function that operates in a similar way to fabtools vagrant.

    Example usage - first add this to your fabfile.py

        from fabgis.docker import docker

    Then use docker instead of a -H hostname directive e.g.::

        fab docker setup_qgis2

    This command makes no attempt to create the docker container, so will fail
    if there is not a fabgis.container.id file in the current working
    directory that contains the id of a valid container.


    .. todo:: Behaviour when more than one host is passed to fabric is still
        not defined or tested.
    """
    host_name = run('hostname')
    fastprint(green(
        'Connected on %s preparing ssh tunnel though this gateway\n' %
        host_name))
    container_id_file = 'fabgis.container.id'
    if exists(container_id_file):
        container_id = run('cat %s' % container_id_file)
    else:
        abort(red('Docker fabgis.container.id file not found.\n'))

    mappings = get_docker_port_mappings(container_id)

    ssh_port = mappings[22]
    env.port = ssh_port
    env.user = 'root'
    fastprint(env)
    host_name = run('hostname')
    fastprint(green(
        'Connected on %s via ssh tunnel though this gateway\n' % host_name))


@task
def setup_docker(force=False):
    """Setup docker on the target host.

    :param force: Whether to continue with installation even if
        docker already appears to be installed. Defaults to False.
    :type force: bool
    """
    fastprint(yellow('Setting up docker on host: %s\n' % env.host))
    if is_installed('lxc-docker'):
        fastprint(green(
            'This system already appears to have docker installed on it\n'))
    else:
        version = run('uname -r')
        if '3.2' in version:
            # LTS 3.2 version is too old so we install a backported one
            # see http://docs.docker.io/en/latest/installation/ubuntulinux/
            # #ubuntu-precise-12-04-lts-64-bit
            fastprint(red('Upgrading kernel to 3.8!\n'))
            response = prompt('Do you wish to continue? y/n :')
            if response != 'y':
                fastprint(red('Docker install aborted by user.\n'))
                return
            fastprint(blue('Ok upgrading kernel.'))
            require_packages([
                'linux-image-generic-lts-raring',
                'linux-headers-generic-lts-raring'])
            fastprint(red('\nWe need to reboot the system now!\n'))
            response = prompt('Do you wish to continue? y/n :')
            if response is not None:
                reboot()
        else:
            require_package('linux-image-extra-%s' % version)
        require_ppa('ppa:dotcloud/lxc-docker')
        apt_get_update()
        require_packages([
            'software-properties-common',
            'lxc-docker'])
    # Ensure ufw forwards traffic.
    # http://docs.docker.io/en/latest/installation/ubuntulinux/#ufw
    sed(
        '/etc/default/ufw',
        'DEFAULT_FORWARD_POLICY="DROP"',
        'DEFAULT_FORWARD_POLICY="ACCEPT"',
        use_sudo=False)
    setup_docker_image()
    setup_docker_user()


@task
def setup_docker_group():
    """Setup the docker group so that any docker group members dont need sudo.

    Once a user is in the docker group they can issue docker commands without
    being a sudo user.
    """
    if not contains('/etc/group', 'docker'):
        sudo('groupadd docker')
        sudo('service docker restart')


@task
def setup_docker_user(user=None):
    """Setup a user with rights to run docker after their next relogin.

    :param user: The user that should be granted docker rights. If none,
        env.user will be assumed.
    :type user: str
    """
    setup_docker_group()
    if user is None:
        user = env.user
    sudo('usermod -a -G docker %s' % user)
    fastprint(red(
        'User %s will be able to run docker without sudo on their '
        'next log in.' % user))


@task
def setup_docker_image():
    """Set up the default docker image to be used in fabgis deployments.

    We base our container on dhrp/sshd image.

    After running this task you will have an image called fabgis/sshd with
    a root password of your choosing.

    On first installation we need to set the root password that will be
    used as default for this base image.
    """

    fastprint(blue(
        'Downloading the dhrp/sshd base image - may take a while.\n'))

    container_id = create_docker_container('dhrp/sshd')

    fastprint(blue(
        'setting the password in the container\n'))
    fastprint(green(
        'Enter the password "screencast" for the first root prompt below.\n'))
    fastprint(green(
        'Enter a password of YOUR OWN CHOOSING at the second prompt below.\n'))
    fastprint(green(
        'Then re-enter your own password at the confirmation prompt.\n'))
    fastprint(red(
        'Note that the password will be echoed to the console in the clear'
        '.\n'))

    mappings = get_docker_port_mappings(container_id)
    run('ssh root@localhost -p %i passwd' % mappings[22])

    fastprint(green(
        'Committing the change to a new default container fabgis/sshd\n'))

    sudo('sudo docker commit %s fabgis/sshd' % container_id)

    fastprint(red(
        'Deleting the temporary container we started during setup.\n'))

    sudo('docker kill %s' % container_id)

    fastprint(green('Installation of docker completed ok!\n'))


@task
def get_docker_port_mappings(container_id):
    """Given the id of a container, get the ports mapped for that container.

    This is basically a wrapper for docker ps that parses the output and
    returns the ports that are mapped as a dict where the keys are the
    container internal ports and the values are the container external ports.

    :param container_id: ID for the container to obtain port mappings for.
    :type container_id: str

    e.g. for this container::

        d3caf337bfc1     dhrp/sshd:latest    /usr/sbin/sshd -D   8 minutes \
        ago       Up 8 minutes        49171->22, 49172->8000

    The following dictionary would be returned::

        {
            22: 49171,
            8000, 49172
        }
    """
    ports = sudo(
        'docker ps | grep %s | awk '
        '\'{ s = ""; for (i = 11; i <= NF; i++) s = s $i " "; print s }\''
        % container_id)

    #fastprint('Ports: %s\n' % str(ports))

    if ', ' in ports:
        tokens = ports.split(', ')
    else:
        tokens = [ports]

    mappings = {}

    #fastprint('Tokens: %s\n' % str(tokens))
    for token in tokens:
        parts = token.split('->')
        #fastprint('Parts: %s\n' % str(parts))
        host_port = int(parts[0])
        container_port = int(parts[1])
        mappings[container_port] = host_port

    fastprint('Mappings: %s\n' % str(mappings))
    return mappings


@task
def create_docker_container(image='fabgis/sshd'):
    """Create a docker container using the provided base image.

    :param image: The image name to be used as a basis for this container.
    :type image: str

    :returns: The id of the created container.
    :rtype: str
    """
    # Id of container will be written to fabgis.container.id

    container_id_file = 'fabgis.container.id'
    if exists(container_id_file):
        sudo('rm %s' % container_id_file)

    sudo(
        'docker run -cidfile=%s -d -p 22 -p 80 %s /usr/sbin/sshd -D' %
        (container_id_file, image))

    # now get the id back - this is more reliable than reading from stdout
    container_id = run('cat %s' % container_id_file)

    return container_id


@task
def current_docker_container():
    """Return the current docker container id as stored in fabgis.container.id.

    :returns: The value in fabgis.container.id. If the container does not exist
        None is returned.
    :rtype: str, None
    """
    container_id_file = 'fabgis.container.id'
    if exists(container_id_file):
        return run('cat %s' % container_id_file)
    else:
        return None


@task
def allow_docker_through_ufw():
    """Allow docker networking to communicate out through UFW.


         TODO Implement this task fully


        http://stackoverflow.com/questions/17394241/my-firewall-is-blocking
        -network-connections-from-the-docker-container-to-outside
    """
    before_rules = """
    # docker rules to enable external network access from the container
    # forward traffic accross the bridge
    -A ufw - before - forward - i  docker0 - j  ACCEPT
    -A ufw - before - forward - i  testbr0 - j  ACCEPT
    -A ufw - before - forward - m  state - -state
    RELATED, ESTABLISHED - j ACCEPT"""

    after_rules = """
    *nat
    :POSTROUTING ACCEPT [0:0]
    -A POSTROUTING -s 172.16.42.0/8 -o eth0 -j MASQUERADE
    # don't delete the 'COMMIT' line or these rules won't be processed
    COMMIT
    """



# coding=utf-8
"""Tools for deployment of tilemill.

.. note:: Usually you should run tilemill on your local host and prepare your
    mbtiles, then use tilestream (https://github.com/mapbox/tilestream) to
    host the resulting mbtiles output.

.. seealso:: :file:`tilestream.py`

"""
import fabtools
from fabric.api import fastprint, task, sudo, run
from .common import setup_env
from .system import get_ip_address


@task
def add_developmentseed_ppa():
    """Ensure we have development seed ppa (makers of mapbox, tilemill etc.."""
    fabtools.deb.update_index(quiet=True)
    fabtools.require.deb.ppa(
        #'ppa:developmentseed/mapbox', auto_yes=True)
        'ppa:developmentseed/mapbox')


@task
def setup_tilemill():
    """Set up tile mill - see http://www.mapbox.com/tilemill/ ."""
    # Note raring seems not to be supported yet...
    setup_env()
    add_developmentseed_ppa()
    fabtools.require.deb.package('tilemill')
    fabtools.require.deb.package('libmapnik')
    # TODO: switch to using nodeenv
    # SEE: https://github.com/ekalinin/nodeenv
    fabtools.require.deb.package('nodejs')
    fastprint('Now you can log in and use tilemill like this:')
    fastprint('vagrant ssh -- -X')
    fastprint('/usr/bin/nodejs /usr/share/tilemill/index.js')
    fastprint('Or use the start tilemill task and open your')
    fastprint('browser at the url provided.')


@task
def start_tilemill():
    """Start the tilemill service - ensure it is installed first."""
    #sudo('start tilemill')
    # Note port forward seems not to work well
    # Using this config on the vhost:
    #{
    #  'listenHost': '0.0.0.0',
    #  'coreUrl': '192.168.1.115:20009',
    #  'tileUrl': '192.168.1.115:20008',
    #  'files': '/usr/share/mapbox',
    #  'server': true
    #}
    # Worked, accessible from http://192.168.1.115:20009/
    # More reliable way (blocking process, ctl-c to kill)
    host_ip = get_ip_address()
    fastprint(host_ip)
    command = (
        '/usr/bin/nodejs /usr/share/tilemill/index.js '
        '--server=true '
        '--listenHost=0.0.0.0 '
        '--coreUrl=%s:20009 '
        '--tileUrl=%s:20008' % (host_ip, host_ip))
    run(command)
    fastprint('Tilemill is running - point your browser at:')
    fastprint('http://%s:20009' % host_ip)
    fastprint('to use it.')



# coding=utf-8
"""General utilities."""
import os
from fabric.api import cd, sudo
from fabric.contrib.files import contains, append


def append_if_not_present(filename, text, use_sudo=False):
    """Append to a file if an equivalent line is not already there.
    :param filename: Name of file to append to.
    :type filename: str

    :param text: Text to append.
    :type text: str

    :param use_sudo: Run the command as sudo
    :type use_sudo: bool
    """
    if not contains(filename, text):
        append(filename, text, use_sudo=use_sudo)


def replace_tokens(conf_file, tokens):
    """Deprecated: prepare a template config file by replacing its tokens.

    :param conf_file: Either a full path to a conf file name or just the
        file name. In the latter case, it assumes the file is then in the
        current working directory. It the file name ends in '.templ',
        a copy will be made and the takens replaced in the copy. Otherwise
        the original file will be manipulated.
    :type conf_file: str

    :param tokens: A dictionary of key-values that should be replaced
        in the conf file.
    :type tokens: dic

    :returns: Path to the replaced file.
    :rtype: str

    Example tokens::

        my_tokens = {
            'SERVERNAME': env.doc_site_name,  # Web Url e.g. foo.com
            'WEBMASTER': 'werner@linfiniti.com',  # email of web master
            'DOCUMENTROOT': webdir,  # Content root .e.g. /var/www
            'SITENAME': sitename,  # Choosen name of jenkins 'root'
        }

    .. deprecated:: You should use fabric.contrib.files.upload_template rather.
    """

    if '.templ' == conf_file[-6:]:
        templ_file = conf_file
        conf_file = conf_file.replace('.templ', '')
        sudo(
            'cp %(templ_file)s %(conf_file)s' % {
            'templ_file': templ_file,
            'conf_file': conf_file})

    base_path, file_name = os.path.split(conf_file)
    if base_path is not '':
        # The file is not in the current working dir.
        with cd(base_path):
            for key, value in tokens.iteritems():
                sudo('sed -i.bak -r -e "s/\[%s\]/%s/g" %s' % (
                    key, value, file_name))
            sudo('rm %s.bak' % file_name)
    else:
        # filename only, not full path - assumes the current working dir is
        # the same as where the conf file is located
        for key, value in tokens.iteritems():
            sudo('sed -i.bak -r -e "s/\[%s\]/%s/g" %s' % (
                key, value, file_name))
            sudo('rm %s.bak' % file_name)
    return conf_file



# coding=utf-8
"""
Mapserver related tasks.
------------------------
"""

import os
from fabric.api import task, fastprint
from fabric.colors import cyan
from fabric.contrib.files import contains, append
from fabtools import fabtools


@task
def setup_mapserver():
    """Install UMN Mapserver from apt and add 900913 CRS to proj epsg file."""
    fastprint(cyan('Setting up UMN Mapserver\n'))
    # Clone and replace tokens in mapserver map file
    # Clone and replace tokens in mapserver conf
    fabtools.require.deb.package('cgi-mapserver')
    # We also need to append 900913 epsg code to the proj epsg list
    epsg_path = '/usr/share/proj/epsg'
    epsg_code = (
        '<900913> +proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_'
        '0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +no_defs')
    epsg_id = '900913'
    if not contains(epsg_path, epsg_id):
        append(epsg_path, epsg_code, use_sudo=True)



# coding=utf-8
"""Tasks for setting up InaSAFE."""

from fabric.api import task, fastprint
from fabric.colors import blue, green
import fabtools


@task
def setup_inasafe():
    """Setup requirements for InaSAFE."""
    fastprint(blue('Setting up InaSAFE dependencies\n'))
    fabtools.require.deb.package('pep8')
    fabtools.require.deb.package('pylint')
    fabtools.require.deb.package('python-nose')
    fabtools.require.deb.package('python-nosexcover')
    fabtools.require.deb.package('python-pip')
    fabtools.require.deb.package('python-numpy')
    fabtools.require.deb.package('python-qt4')
    fabtools.require.deb.package('python-nose')
    fastprint(green('Setting up InaSAFE dependencies completed.\n'))



# coding=utf-8
"""Tasks for install and setup of java."""

from fabric.api import run, sudo, task, fastprint, env
from fabric.colors import red, green, blue, yellow
from fabtools.deb import update_index as apt_get_update
from fabtools.require.deb import ppa as require_ppa
from fabtools.require.deb import package as require_package


@task
def install_oracle_jdk():
    """Install the official oracle jdk."""
    fastprint(yellow('Setting up oracle java on host: %s' % env.host))
    require_ppa('ppa:webupd8team/java')
    apt_get_update()
    require_package('software-properties-common')
    sudo(
        'echo oracle-java7-installer shared/accepted-oracle-license-v1-1 '
        'select true | debconf-set-selections')
    require_package('oracle-java7-installer')
    fastprint(green('Installation of oracle java completed ok!'))




# coding=utf-8
"""Postgres related fabric tasks and helpers."""
import os
import fabtools
from fabric.contrib.files import exists
from fabtools.postgres import create_user
from fabric.api import run, cd, env, task, sudo, get, put
from .common import setup_env, show_environment, add_ubuntugis_ppa
from .utilities import replace_tokens


@task
def require_postgres_user(user, password='', createdb=False):
    """ Require a postgres username to create a database
    :param user: username
    :param password: password
    :param createdb:  (default=False)
    :return:
    """
    #sudo('apt-get upgrade')
    # wsgi user needs pg access to the db
    if not fabtools.postgres.user_exists(user):
        # noinspection PyArgumentEqualDefault
        fabtools.postgres.create_user(
            name=user,
            password=password,
            superuser=False,
            createdb=createdb,
            createrole=False,
            inherit=True,
            login=True,
            connection_limit=None,
            encrypted_password=False)


def setup_postgres_superuser(user, password=''):
    """Create a super user for postgresql.

    :param user: User name for new super user.
    :param password:  Password for new user.
    """
    if not fabtools.postgres.user_exists(user):
        fabtools.postgres.create_user(
            user,
            password=password,
            createdb=True,
            createrole=True,
            superuser=True,
            connection_limit=20)


@task
def create_postgis_2_template():
    """Create the postgis 2 template db."""
    if not fabtools.postgres.database_exists('template_postgis'):
        setup_postgres_superuser(env.user)
        # noinspection PyArgumentEqualDefault
        fabtools.require.postgres.database(
            'template_postgis',
            owner='%s' % env.user,
            encoding='UTF8')
        sql = ('UPDATE pg_database SET datistemplate = TRUE WHERE datname = '
               '\'template_postgis\';')
        run('psql template1 -c "%s"' % sql)
        sql_path = '/usr/share/postgresql/9.1/contrib/postgis-2.0/'
        run('psql template_postgis -f %s/postgis.sql' % sql_path)
        run('psql template_postgis -f %s/spatial_ref_sys.sql' % sql_path)
        grant_sql = 'GRANT ALL ON geometry_columns TO PUBLIC;'
        run('psql template_postgis -c "%s"' % grant_sql)
        grant_sql = 'GRANT ALL ON geography_columns TO PUBLIC;'
        run('psql template_postgis -c "%s"' % grant_sql)
        grant_sql = 'GRANT ALL ON spatial_ref_sys TO PUBLIC;'
        run('psql template_postgis -c "%s"' % grant_sql)


@task
def setup_postgis_2():
    """Set up postgis 2.0 from packages in ubuntugis."""
    add_ubuntugis_ppa()
    fabtools.require.deb.package('build-essential')
    fabtools.require.deb.package('postgresql-9.1-postgis-2.0')
    fabtools.require.deb.package('postgresql-9.1-postgis-2.0-scripts')
    fabtools.require.deb.package('postgresql-server-dev-all')
    create_postgis_2_template()


@task
def setup_postgis_1_5():
    """Set up postgis.

    You can call this multiple times without it actually installing all over
    again each time since it checks for the presence of pgis first.

    We build from source because we want 1.5"""
    setup_env()

    pg_file = '/usr/share/postgresql/9.1/contrib/postgis-1.5/postgis.sql'
    if not fabtools.files.is_file(pg_file):
        add_ubuntugis_ppa()
        fabtools.require.deb.package('postgresql-server-dev-all')
        fabtools.require.deb.package('build-essential')

        # Note - no postgis installation from package as we want to build 1.5
        # from source
        fabtools.require.postgres.server()

        # Now get and install postgis 1.5 if needed

        fabtools.require.deb.package('libxml2-dev')
        fabtools.require.deb.package('libgeos-dev')
        fabtools.require.deb.package('libgdal1-dev')
        fabtools.require.deb.package('libproj-dev')
        source_url = ('http://download.osgeo.org/postgis/source/'
                      'postgis-1.5.8.tar.gz')
        source = 'postgis-1.5.8'
        if not fabtools.files.is_file('%s.tar.gz' % source):
            run('wget %s' % source_url)
            run('tar xfz %s.tar.gz' % source)
        with cd(source):
            run('./configure')
            run('make')
            sudo('make install')

    create_postgis_1_5_template()


@task
def create_postgis_1_5_template():
    """Create the postgis template db."""
    if not fabtools.postgres.database_exists('template_postgis'):
        setup_postgres_superuser(env.user)
        # noinspection PyArgumentEqualDefault
        fabtools.require.postgres.database(
            'template_postgis',
            owner='%s' % env.user,
            encoding='UTF8')
        sql = ('UPDATE pg_database SET datistemplate = TRUE WHERE datname = '
               '\'template_postgis\';')
        run('psql template1 -c "%s"' % sql)
        run('psql template_postgis -f /usr/share/postgresql/'
            '9.1/contrib/postgis-1.5/postgis.sql')
        run('psql template_postgis -f /usr/share/postgresql/9'
            '.1/contrib/postgis-1.5/spatial_ref_sys.sql')
        grant_sql = 'GRANT ALL ON geometry_columns TO PUBLIC;'
        run('psql template_postgis -c "%s"' % grant_sql)
        grant_sql = 'GRANT ALL ON geography_columns TO PUBLIC;'
        run('psql template_postgis -c "%s"' % grant_sql)
        grant_sql = 'GRANT ALL ON spatial_ref_sys TO PUBLIC;'
        run('psql template_postgis -c "%s"' % grant_sql)


@task
def create_postgis_1_5_db(dbname, user):
    """Create a postgis database.
    :param dbname: Name of database to create.
    :param user: User who should own the created db
    """
    setup_postgis_1_5()
    setup_postgres_superuser(env.user)
    require_postgres_user(user)
    fabtools.require.postgres.database(
        '%s' % dbname, owner='%s' % user, template='template_postgis')

    grant_sql = 'GRANT ALL ON schema public to %s;' % user
    # assumption is env.repo_alias is also database name
    run('psql %s -c "%s"' % (dbname, grant_sql))
    grant_sql = (
        'GRANT ALL ON ALL TABLES IN schema public to %s;' % user)
    # assumption is env.repo_alias is also database name
    run('psql %s -c "%s"' % (dbname, grant_sql))
    grant_sql = (
        'GRANT ALL ON ALL SEQUENCES IN schema public to %s;' % user)
    run('psql %s -c "%s"' % (dbname, grant_sql))


@task
def get_postgres_dump(dbname, ignore_permissions=False, file_name=None):
    """Get a dump of the database from the server.

    :param dbname: name of the database to restore the dump into.
    :type dbname: str

    :param ignore_permissions: whether permissions in the created dump
        should be preserved.
    :type ignore_permissions: bool (default False)

    :param file_name: optional file name for the dump. The file name should
        exclude any path. If file_name is ommitted, the dump will be written to
        fabgis_resources/sql/dumps/<dbname>->date>.dmp
        where date is in the form dd-mm-yyyy. This is the default naming
        convention used by the :func:`restore_postgres_dump` function below.
    :type file_name: str
    """
    setup_env()

    if file_name is None or file_name == '':
        date = run('date +%d-%B-%Y')
        my_file = '%s-%s.dmp' % (dbname, date)
    else:
        my_file = os.path.split(file_name)[1]
        put(file_name, '/tmp/%s' % my_file)

    if not ignore_permissions:
        extra_args = ''
    else:
        extra_args = '-x -O'

    run('pg_dump %s -Fc -f /tmp/%s %s' % (extra_args, my_file, dbname))
    get('/tmp/%s' % my_file, 'fabgis_resources/sql/dumps/%s' % my_file)


@task
def restore_postgres_dump(
        dbname,
        user=None,
        password='',
        ignore_permissions=False,
        file_name=None):
    """Upload dump to host, remove existing db, recreate then restore dump.

    :param dbname: name of the database to restore the dump into.
    :type dbname: str

    :param user: user that the db should be restored for. The db user
        will be used when restoring the db and the user will be created first
        if needed.
    :type user: str

    :param password: password to use for db connection - defaults to ''
    :type password: str

    :param ignore_permissions: whether permissions in the restored dump
        should be retained.
    :type ignore_permissions: bool (default False)

    :param file_name: optional file name for the dump. If ommitted,
        the dump will be assumed to exist in
        fabgis_resources/sql/dumps/<dbname>->date>.dmp
        where date is in the form dd-mm-yyyy. This is the default naming
        convention used by the :func:`get_postgres_dump` function above.
    :type file_name: str
    """
    setup_env()
    if user is None:
        user = env.fg.user
    show_environment()
    require_postgres_user(user, password=password)
    if file_name is None or file_name == '':
        date = run('date +%d-%B-%Y')
        my_file = '%s-%s.dmp' % (dbname, date)
        put('fabgis_resources/sql/dumps/%s' % my_file, '/tmp/%s' % my_file)
    else:
        my_file = os.path.split(file_name)[1]
        put(file_name, '/tmp/%s' % my_file)

    if fabtools.postgres.database_exists(dbname):
        run('dropdb %s' % dbname)

    # noinspection PyArgumentEqualDefault
    fabtools.require.postgres.database(
        '%s' % dbname,
        owner='%s' % user,
        template='template_postgis',
        encoding='UTF8')

    if not ignore_permissions:
        extra_args = ''
    else:
        extra_args = '-x -O'
    run('pg_restore %s /tmp/%s | psql %s' % (extra_args, my_file, dbname))


@task
def setup_nightly_backups():
    """Setup nightly backups for all postgresql databases.

    The template script :file:`fabgis_resources/server_config/cron/pg_backups`
    will place the last 21 days of backups in the remote user's Dropbox
    directory and will maintain 6 months of backups in their home directory.

    .. seealso:: fabgis.dropbox for help on setting up dropbox on your server.
    """
    setup_env()
    setup_postgres_superuser(env.fg.user)
    with cd('/etc/cron.daily/'):
        if exists('pg_backups'):
            sudo('rm pg_backups')

        local_dir = os.path.dirname(__file__)
        local_file = os.path.abspath(os.path.join(
            local_dir,
            '..',
            '../fabgis_resources',
            'server_config',
            'cron',
            'pg_backups.templ'))
        put(local_file,
            '/etc/cron.daily/pg_backups',
            use_sudo=True)

        my_tokens = {'USER': env.fg.user, }
        replace_tokens('/etc/cron.daily/pg_backups', my_tokens)

    sudo('chmod +x /etc/cron.daily/pg_backups')
    # Run once to verify it works
    sudo('/etc/cron.daily/pg_backups')



import os
import time
from fabric.contrib.files import contains, exists
from fabric.api import run, cd, env, task, sudo, put
import fabtools
from .utilities import append_if_not_present, replace_tokens
from .common import setup_env


@task
def add_jenkins_repository():
    """Add the Jenkins latest repository"""
    jenkins_url = 'http://pkg.jenkins-ci.org/'
    jenkins_key = jenkins_url + 'debian/jenkins-ci.org.key'
    jenkins_repo = jenkins_url + 'debian binary/'
    jenkins_apt_file = '/etc/apt/sources.list.d/jenkins-repository.list'

    run('wget -q -O - %s | sudo apt-key add -' % jenkins_key)

    if not exists(jenkins_apt_file):
        sudo('touch %s' % jenkins_apt_file)
    append_if_not_present(
        jenkins_apt_file, 'deb ' + jenkins_repo, use_sudo=True)


def configure_jenkins(repo):
    """
    We have to update the json file if we want to switch between distribution
    provided jenkins and upstream jenkins
    """
    command1 = 'curl -L http://updates.jenkins-ci.org/update-center.json'
    command2 = 'sed "1d;$d" > /var/lib/jenkins/updates/default.json'
    jenkins = 'jenkins'
    sudo('mkdir -p /var/lib/jenkins/updates', user=jenkins)
    sudo('%s > /var/lib/jenkins/updates/default.json' % command1,
         user=jenkins)
    sudo('%s | %s' % (command1, command2), user=jenkins)
    deploy_jenkins_plugins(repo=repo)


@task
def deploy_jenkins_plugins(repo):
    """
    Deploying plugins with jenkins-cli seems not to be reliable
    For now if the plugins are not getting installed automatically
    we have to go to the webinterface and  install at least github and xvfb
    plugin manually before calling setup_jenkins_jobs
    :param repo: repository to use
    :return:
    """
    webpath = 'http://updates.jenkins-ci.org'
    distribution = repo
    if distribution == 'upstream':
        command = 'java -jar /var/cache/jenkins/war/WEB-INF/jenkins-cli.jar'

        run('%s -s http://localhost:8080 install-plugin '
            '%s/download/plugins/github/1.5/github.hpi'
            % (command, webpath))
        run('%s -s http://localhost:8080 install-plugin %s/latest/xvfb.hpi'
            % (command, webpath))
        run('%s -s http://localhost:8080 install-plugin '
            '%s/latest/violation-columns.hpi'
            % (command, webpath))
        run('%s -s http://localhost:8080 install-plugin '
            '%s/latest/statusmonitor.hpi'
            % (command, webpath))
        run('%s -s http://localhost:8080 install-plugin %s/latest/sounds.hpi'
            % (command, webpath))
        run('%s -s http://localhost:8080 install-plugin '
            '%s/latest/maven-plugin.hpi'
            % (command, webpath))
        run('%s -s http://localhost:8080 install-plugin '
            '%s/latest/covcomplplot.hpi'
            % (command, webpath))
        run('%s -s http://localhost:8080 install-plugin %s/latest/cobertura.hpi'
            % (command, webpath))
        run('%s -s http://localhost:8080 install-plugin '
            '%s/latest/dashboard-view.hpi'
            % (command, webpath))
        run('%s -s http://localhost:8080 install-plugin %s/latest/sloccount.hpi'
            % (command, webpath))
        sudo('sudo service jenkins restart')
    else:
        command = 'jenkins-cli'
        # selecting a specific version
        run('%s -s http://localhost:8080 install-plugin '
            '%s/download/plugins/github/1.6/github.hpi' % (command, webpath))
        # just install the latest ones with jenkins-cli
        # running jenkins-cli that way not possible with latest jenkins
        run('%s -s http://localhost:8080 install-plugin xvfb' % command)
        run('%s -s http://localhost:8080 install-plugin violations' % command)
        run('%s -s http://localhost:8080 install-plugin statusmonitor' %
            command)
        run('%s -s http://localhost:8080 install-plugin sounds' % command)
        run('%s -s http://localhost:8080 install-plugin %s/maven-plugin/1.480'
            '.3/maven-plugin' % (command, webpath))
        run('%s -s http://localhost:8080 install-plugin covcomplplot' % command)
        run('%s -s http://localhost:8080 install-plugin cobertura' % command)
        run('%s -s http://localhost:8080 install-plugin dashboard-view' %
            command)
        run('%s -s http://localhost:8080 install-plugin sloccount' % command)
        sudo('sudo service jenkins restart')


@task
def install_jenkins(use_upstream_repo=False):
    """Add latest jenkins with adding jenkins project repository.

    .. note:: If you install only the distribution packaged version,
        some of the plugins listed in the configure_jenkins target may not
        install.

    Args:
        use_upstream_repo - bool: (defaults to False). Whether to use the
        official jenkins repo or not. In the case of False,
        the distribution repo version will be used instead.

    Example:
        Using the upstream jenkins repo:
        fab -H localhost fabgis.fabgis.install_jenkins:use_upstream_repo=True

    """
    if use_upstream_repo:
        repo = 'upstream'
        if fabtools.deb.is_installed('jenkins'):
            fabtools.deb.uninstall(['jenkins-common',
                                    'jenkins-cli',
                                    'libjenkins-remoting-java'], purge=True)
        add_jenkins_repository()
        fabtools.deb.update_index(quiet=True)
        sudo('apt-get install jenkins')
        #Jenkins needs time to come up
        time.sleep(5)
        configure_jenkins(repo)
    else:
        repo = 'distribution'
        if os.path.exists('/etc/apt/sources.list.d/jenkins-repository.list'):
            sudo('rm /etc/apt/sources.list.d/jenkins-repository.list')
        fabtools.deb.update_index(quiet=True)
        if fabtools.deb.is_installed('jenkins'):
            fabtools.deb.uninstall(['jenkins-common',
                                    'jenkins-cli',
                                    'libjenkins-remoting-java'], purge=True)
        fabtools.require.deb.package('jenkins')
        fabtools.require.deb.package('jenkins-common')
        fabtools.require.deb.package('jenkins-cli')
        #Jenkins needs time to come up
        time.sleep(5)
        configure_jenkins(repo)


@task
def jenkins_deploy_website(site_url=None, use_upstream_repo=False):
    """
    Initialise jenkins with local proxy if we are not going to use port
    8080
    """
    setup_env()
    fabtools.require.deb.package('apache2')
    sitename = site_url

    if not sitename:
        sitename = 'jenkins'
    else:
        sitename = site_url

    jenkins_apache_conf = ('fabgis.%s.conf' % (sitename))
    jenkins_apache_conf_template = 'fabgis.jenkins.conf.templ'

    with cd('/etc/apache2/sites-available/'):
        if not exists(jenkins_apache_conf):
            local_dir = os.path.dirname(__file__)
            local_file = os.path.abspath(os.path.join(
                local_dir,
                '..',
                '../fabgis_resources',
                'server_config',
                'apache',
                jenkins_apache_conf_template))
            put(local_file,
                '/etc/apache2/sites-available/%s' %
                jenkins_apache_conf_template,
                use_sudo=True)

        my_tokens = {
            'SERVERNAME': env.fg.hostname,  # Web Url e.g. foo.com
            'WEBMASTER': 'werner@linfiniti.com',  # email of web master
            'SITENAME': sitename,  # Choosen name of jenkins 'root'
        }
        replace_tokens(jenkins_apache_conf_template, my_tokens)

    # Add a hosts entry for local testing - only really useful for localhost
    hosts = '/etc/hosts'
    if not contains(hosts, '%s.%s' % (sitename, env.fg.hostname)):
        append_if_not_present(hosts,
                              '127.0.1.1 %s.%s' % (sitename, env.fg.hostname),
                              use_sudo=True)
        append_if_not_present(hosts,
                              '127.0.0.1 %s.localhost' % env.fg.hostname,
                              use_sudo=True)
        # For doing Reverse Proxy we need to enable 2 apache modules
    sudo('a2enmod proxy')
    sudo('a2enmod proxy_http')
    sudo('a2ensite %s' % jenkins_apache_conf)
    sudo('service apache2 reload')



# coding=utf-8
"""Motion Web Cam
==================


This module provides tools for setting up the motion web cam daemon.

http://www.lavrsen.dk/foswiki/bin/view/Motion/WebHome

"""
import os
from fabgis.fabgis import sed
from fabric.api import cd, task, sudo, fastprint, run
from fabric.colors import green
from fabric.contrib.files import upload_template
from fabtools import require
from .common import setup_env


@task
def setup_motion(email):
    """Set up the motion web cam daemon.

    :param email: Email address to send motion alerts to.
    :type email: str

    """
    require.deb.package('motion')
    require.deb.package('mpack')
    sed(
        '/etc/default/motion',
        'start_motion_daemon=no',
        'start_motion_daemon=yes',
        use_sudo=True)
    sed(
        '/etc/default/motion',
        '; on_movie_end value',
        ('on_movie_end mpack -s \'[Motion Alert] %%Y-%%m-%%d %%H:%%M:%%S\' '
         '/tmp/motion/%f %s' % email),
        use_sudo=True)
    require.service.restarted('motion')


@task
def setup_apache(
        site_name,
        web_root,
        template_dir=None,
        **kwargs):
    """Set up the apache server for this site.

    :param site_name: Name of the site e.g. cam.linfiniti.com. Should be a
        single word with only alpha characters in it.
    :type site_name: str

    :param web_root: Directory where the content lives.
    :type web_root: str

    :param template_dir: Directory where the template files live. If none
        will default to ``resources/server_config/apache``. Must be a
        relative path to the fabfile you are running.
    :type template_dir: str

    :param kwargs: Any extra keyword arguments that should be appended to the
        token list that will be used when rendering the apache config template.
        Use this to pass in sensitive data such as passwords.
    :type kwargs: dict

    :returns: Path to the apache conf file.
    :rtype: str
    """
    setup_env()
    # Ensure we have a mailserver setup for our domain
    # Note that you may have problems if you intend to run more than one
    # site from the same server
    require.postfix.server(site_name)

    # Clone and replace tokens in apache conf
    if template_dir is None:
        template_dir = 'resources/server_config/apache/'
    filename = '%s.apache.conf.templ' % site_name
    template_path = os.path.join(template_dir, filename)
    fastprint(green('Using %s for template' % template_path))

    context = {
        'escaped_server_name': site_name.replace('.', '\.'),
        'server_name': site_name,
        'web_root': web_root,
        'site_name': site_name}
    context.update(kwargs)  # merge in any params passed in to this function
    destination = '/etc/apache2/sites-available/%s.apache.conf' % site_name
    fastprint(context)

    upload_template(
        template_path,
        destination,
        context=context,
        use_sudo=True)

    require.deb.package('apache2')
    sudo('a2dissite default')
    sudo('a2ensite %s.apache.conf' % site_name)
    # Check if apache configs are ok - script will abort if not ok
    sudo('/usr/sbin/apache2ctl configtest')
    require.service.restarted('apache2')
    return destination



# coding=utf-8
"""Set up a Hetzner server from scratch."""

from fabric.api import task, run, local


@task
def hetzner_setup(host_name):
    """Provision a hetzner server with software raid.

    :param host_name: The fully qualified domain name to use for the host.
    :type host_name: str

    .. example::

        fab -H mts2 hetzner_setup:host_name=mappingthesacred.org

    If you have an ssh agent running you may be prompted for your password::

        [mts2] Passphrase for private key:

    Enter any text to skip it repeatedly until you get prompted for an
    interactive password.

    Example session output::

        out:                 Hetzner Online AG - installimage
        out:
        out:   Your server will be installed now, this will take some minutes
        out:              You can abort at any time with CTRL+C ...
        out:
        out:          :  Reading configuration               done
        out:    1/14  :  Deleting partitions                 done
        out:    2/14  :  Test partition size                 done
        out:    3/14  :  Creating partitions and /etc/fstab  done
        out:    4/14  :  Creating software RAID level 1      done
        out:    5/14  :  Formatting partitions
        out:          :    formatting /dev/md/0 with ext3    done
        out:          :    formatting /dev/md/1 with ext4    done
        out:    6/14  :  Mounting partitions                 done
        out:    7/14  :  Extracting image (local)            done
        out:    8/14  :  Setting up network for eth0         done
        out:    9/14  :  Executing additional commands
        out:          :    Generating new SSH keys           done
        out:          :    Generating mdadm config           done
        out:          :    Generating ramdisk                done
        out:          :    Generating ntp config             done
        out:          :    Setting hostname                  done
        out:   10/14  :  Setting up miscellaneous files      done
        out:   11/14  :  Setting root password               done
        out:   12/14  :  Installing bootloader grub          done
        out:   13/14  :  Running some ubuntu specific functi done
        out:   14/14  :  Clearing log files                  done
        out:
        out:                   INSTALLATION COMPLETE
        out:    You can now reboot and log in to your new system with
        out:   the same password as you logged in to the rescue system.
        out:
        out:


    """
    options = {
        'install_image_command': '/root/.oldroot/nfs/install/installimage',
        'boot_partition_size': '4G',
        'raid_members': 'sda,sdb',
        'image': 'Ubuntu-1204-precise-64-minimal.tar.gz',
        'host_name': host_name}

    setup_command = (
        '%(install_image_command)s -a -p /boot:ext3:%(boot_partition_size)s,'
        '/:ext4:all -r yes -l 1 -b grub -d %(raid_members)s -i '
        '/root/.oldroot/nfs/install/../images/%(image)s '
        '-n %(host_name)s' % options)

    run(setup_command)
    ip = run('hostname -I')  #
    ip = ip.split(' ')[0]
    run('reboot')
    # Uncache the servers key as it will have changed after re-imaging
    command = 'ssh - keygen - f  "~/.ssh/known_hosts" - R %s' % ip
    local(command)



# coding=utf-8
"""Module for gdal related tasks"""
import fabtools
from fabric.contrib.files import exists, append
from fabric.api import fastprint, run, cd, env, task, sudo, settings

from .common import add_ubuntugis_ppa, setup_env
from .system import setup_ccache
from .proj4 import build_proj4


@task
def build_gdal(with_ecw=False, with_mrsid=False):
    """Clone or update GDAL from svn then build it.

    :param with_ecw: Whether to build with ecw support.
    :type with_ecw: bool

    :param with_mrsid: Whether to build with mrsid support.
    :type with_mrsid: bool

    """
    setup_env()
    add_ubuntugis_ppa()
    fabtools.require.deb.package('subversion')
    fabtools.require.deb.package('build-essential')
    setup_ccache()
    fabtools.require.deb.package('libhdf5-serial-dev')
    fabtools.require.deb.package('libhdf5-7')
    fabtools.require.deb.package('libhdf4g-dev')
    fabtools.require.deb.package('libjpeg62-dev')
    fabtools.require.deb.package('libtiff4-dev')
    fabtools.require.deb.package('python-dev')

    # Note that gdal does not compile against proj4, only uses the .so at
    # runtime
    build_proj4()

    code_base = '%s/cpp' % env.fg.workspace
    code_path = '%s/gdal' % code_base
    if not exists(code_path):
        fastprint('Repo checkout does not exist, creating.')
        run('mkdir -p %s' % code_base)
        with cd(code_base):
            run('svn co %s gdal' % env.fg.gdal_svn_url)
    else:
        fastprint('Repo checkout does exist, updating.')
        with cd(code_path):
            # Get any updates first
            run('svn update')

    flags = (
        '--with-libtiff=internal '
        '--with-geotiff=internal '
        '--with-python '
        '--without-jp2mrs '
        '--with-spatialite '
        '--without-libtool')

    processor_count = run('cat /proc/cpuinfo | grep processor | wc -l')

    # Currently you need to have downloaded the MRSID sdk to remote home dir
    if with_mrsid:
        mrsid_dir = '/usr/local/GeoExpressSDK'
        if not exists(mrsid_dir):
            sudo('tar xfz /home/%s/Geo_DSDK-7.0.0.2167.linux.x86-64.gcc41.'
                 'tar.gz -C /tmp' % env.user)
            sudo('mv /tmp/Geo_DSDK-7.0.0.2167 %s' % mrsid_dir)
        flags += ' --with-mrsid=%s' % mrsid_dir

    # Currently you need to have downloaded the ECW sdk to remote home dir
    if with_ecw:
        ecw_dir = '%s/ecw/libecwj2-3.3' % code_base
        ecw_archive = 'ecw.tar.bz2'
        if not exists(ecw_dir):
            with cd(code_base):
                run('tar xfj ~/%s' % ecw_archive)
        if not exists('/usr/local/include/ECW.h'):
            with cd(ecw_dir):
                run('./configure')
                run('make -j %s' % processor_count)
                sudo('make install')
        flags += ' --with-ecw=/usr/local'

    with cd(code_path):
        # Dont fail if make clean does not work
        with settings(warn_only=True):
            run('make clean')
        run('CXXFLAGS=-fPIC ./configure %s' % flags)
        run('make -j %s' % processor_count)
        sudo('make install')
    # Write to ld path too so libs are loaded nicely
    ld_file = '/etc/ld.so.conf.d/usr_local_lib.conf'
    with settings(warn_only=True):
        sudo('rm %s' % ld_file)
    append(ld_file, '/usr/local/lib', use_sudo=True)
    sudo('ldconfig')



# coding=utf-8
"""Tools for setting up and hardening a system."""

from getpass import getpass

from fabric.api import cd, fastprint, prompt
from fabric.contrib.files import contains, exists, append, sed
from fabric.colors import red
from fabric.api import env, task, sudo, local, reboot
import fabtools
from .utilities import append_if_not_present


def setup_developer_tools():
    """Install various useful tools needed for developers."""
    fabtools.require.deb.package('qtcreator')
    fabtools.require.deb.package('qt4-designer')
    fabtools.require.deb.package('qt4-linguist-tools')


def setup_ccache():
    """Setup ccache."""
    fabtools.require.deb.package('ccache')
    sudo('ln -fs /usr/bin/ccache /usr/local/bin/gcc')
    sudo('ln -fs /usr/bin/ccache /usr/local/bin/g++')
    sudo('ln -fs /usr/bin/ccache /usr/local/bin/cc')


@task
def install_modxsend():
    """
    Download, compile and activate mod_xsendfile
    """
    fabtools.require.deb.package('apache2-threaded-dev')
    with cd('/tmp'):
        sudo('wget https://tn123.org/mod_xsendfile/mod_xsendfile.c')
        sudo('apxs2 -cia mod_xsendfile.c')
    fabtools.require.service.restarted('apache2')


@task
def install_elasticsearch():
    """
    Download and unpack elasticsearch
    """
    sudo(
        'wget https://download.elasticsearch.org/elasticsearch/'
        'elasticsearch/elasticsearch-0.90.2.deb')
    sudo('sudo dpkg -i elasticsearch-0.90.2.deb')
    fabtools.require.service.restarted('elasticsearch')


@task
def create_user(user, password=None):
    """Create a user on the remote system matching the user running this task.

    :param user: User name for the new user.
    :type user: str

    :param password: Password for new user - will prompt interactively if None.
    :type password: str
    """
    if password is None:
        fastprint(red('Please enter a password for the new web user.\n'))
        password = getpass()
    fabtools.require.users.user(user, password=password)
    fabtools.require.users.sudoer(user)


@task
def ssh_copy_id():
    """Copy ssh id from local system to remote.
    .. note:: Does not work on OSX!
    """
    command = 'ssh-copy-id %s' % env.host
    local(command)


@task
def setup_mosh():
    """ Install mosh as a nice and always working kind of replacement to ssh
    """
    mosh_file = '/etc/ufw/applications.d/mosh'
    if not exists(mosh_file):
        sudo('touch %s' % mosh_file)

        append(mosh_file, '[mosh]', use_sudo=True)
        append(mosh_file, ('title=Mobile shell that supports roaming and '
                           'intelligent local echo.'), use_sudo=True)
        append(mosh_file, ('description=The mosh provides alternative remote '
                           'shell that supports roaming and intelligent local '
                           'echo.'), use_sudo=True)
        append(mosh_file, 'ports=60000:61000/udp', use_sudo=True)
    fabtools.require.deb.package('mosh')


def get_ip_address():
    """Get the ip address of the remote host.

    :returns: Ip address of the remote host.
    :rtype: str
    """
    host_ip = sudo(
        "ifconfig eth0 | grep 'inet addr:'| "
        "cut -d: -f2 | awk '{print $1}'")
    return host_ip


@task
def harden(ssh_port=22):
    """Harden the server a little.

    :param ssh_port:

    Warning: We make no claim that this makes your server intruder proof. You
    should always check any system yourself and make sure that it is
    adequately secured.

    .. todo:: Make this work more gracefully if harden has been run previously.

    """
    # Create a user name because after we are done remote login as root will
    # be disabled. Username will match your local user.

    user = prompt('Choose a user name')
    password = prompt('Choose a password for the new user')

    create_user(user, password)
    ssh_copy_id()  # this does not work on OSX
    if not contains('/etc/group', 'admin'):
        sudo('groupadd admin')
    sudo('usermod -a -G admin %s' % user)
    sudo('dpkg-statoverride --update --add root admin 4750 /bin/su')

    fabtools.deb.update_index(quiet=True)

    # Set up ufw and mosh
    fabtools.require.deb.package('ufw')
    setup_mosh()

    sudo('ufw default deny incoming')
    sudo('ufw default allow outgoing')
    sudo('ufw allow 8697')
    sudo('ufw allow http')
    sudo('ufw allow ssh')
    sudo('ufw allow mosh')
    sudo('ufw allow 25')  # mail
    #Irc freenode
    sudo('ufw allow from 127.0.0.1/32 to 78.40.125.4 port 6667')
    sudo('ufw allow from 127.0.0.1/32 to any port 22')
    sudo('ufw allow 443')
    sudo('ufw allow 53/udp')  # dns
    sudo('ufw allow 53/tcp')
    sudo('ufw allow 1053')  # dns client

    sed('/etc/ssh/sshd_config', 'Port 22', 'Port 8697', use_sudo=True)
    sed('/etc/ssh/sshd_config', 'PermitRootLogin yes',
        'PermitRootLogin no', use_sudo=True)
    sed('/etc/ssh/sshd_config', '#PasswordAuthentication yes',
        'PasswordAuthentication no', use_sudo=True)
    sed('/etc/ssh/sshd_config', 'X11Forwarding yes',
        'X11Forwarding no', use_sudo=True)
    sudo('ufw enable')

    append_if_not_present(
        '/etc/ssh/sshd_config', 'Banner /etc/issue.net', use_sudo=True)

    append_if_not_present(
        '/etc/sysctl.conf',
        'net.ipv4.conf.default.rp_filter=1', use_sudo=True)
    append_if_not_present(
        '/etc/sysctl.conf',
        'net.ipv4.conf.setup_env.rp_filter=1', use_sudo=True)
    append_if_not_present(
        '/etc/sysctl.conf',
        'net.ipv4.conf.setup_env.accept_redirects = 0', use_sudo=True)
    append_if_not_present(
        '/etc/sysctl.conf',
        'net.ipv4.conf.setup_env.send_redirects = 0', use_sudo=True)
    append_if_not_present(
        '/etc/sysctl.conf',
        'net.ipv4.conf.setup_env.accept_source_route = 0', use_sudo=True)
    append_if_not_present(
        '/etc/sysctl.conf',
        'net.ipv4.icmp_echo_ignore_broadcasts = 1', use_sudo=True)
    append_if_not_present(
        '/etc/sysctl.conf',
        'net.ipv4.icmp_ignore_bogus_error_responses = 1', use_sudo=True)

    fabtools.require.deb.package('denyhosts')
    # Must come before mailutils
    fabtools.require.postfix.server(env.host)
    fabtools.require.deb.package('mailutils')
    fabtools.require.deb.package('byobu')
    fabtools.service.restart('ssh')

    # Some hints and tips from:
    # http://www.thefanclub.co
    # .za/how-to/how-secure-ubuntu-1204-lts-server-part-1-basics
    secure_tmp = (
        'tmpfs     /dev/shm     tmpfs     defaults,noexec,'
        'nosuid     0     0')
    append_if_not_present('/etc/fstab', secure_tmp, use_sudo=True)

    sysctl = '/etc/sysctl.conf'

    append_if_not_present(
        sysctl, '# IP Spoofing protection', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.conf.setup_env.rp_filter = 1', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.conf.default.rp_filter = 1', use_sudo=True)

    append_if_not_present(
        sysctl, '# Ignore ICMP broadcast requests', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.icmp_echo_ignore_broadcasts = 1', use_sudo=True)

    append_if_not_present(
        sysctl, '# Disable source packet routing', use_sudo=True)
    append_if_not_present(
        sysctl,
        'net.ipv4.conf.setup_env.accept_source_route = 0', use_sudo=True)
    append_if_not_present(
        sysctl,
        'net.ipv6.conf.setup_env.accept_source_route = 0', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.conf.default.accept_source_route = 0', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv6.conf.default.accept_source_route = 0', use_sudo=True)

    append_if_not_present(
        sysctl, '# Ignore send redirects', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.conf.setup_env.send_redirects = 0', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.conf.default.send_redirects = 0', use_sudo=True)

    append_if_not_present(
        sysctl, '# Block SYN attacks', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.tcp_syncookies = 1', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.tcp_max_syn_backlog = 2048', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.tcp_synack_retries = 2', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.tcp_syn_retries = 5', use_sudo=True)

    append_if_not_present(
        sysctl, '# Log Martians', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.conf.setup_env.log_martians = 1', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.icmp_ignore_bogus_error_responses = 1', use_sudo=True)

    append_if_not_present(
        sysctl, '# Ignore ICMP redirects', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.conf.setup_env.accept_redirects = 0', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv6.conf.setup_env.accept_redirects = 0', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.conf.default.accept_redirects = 0', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv6.conf.default.accept_redirects = 0', use_sudo=True)

    append_if_not_present(
        sysctl, '# Ignore Directed pings', use_sudo=True)
    append_if_not_present(
        sysctl, 'net.ipv4.icmp_echo_ignore_all = 1', use_sudo=True)

    sudo('sysctl -p')
    reboot()

    print 'You need to log in and install mailutils yourself as automated ' \
          'installation causes interactive prompting.'
    print 'sudo apt-get install mailutils'


@task
def setup_bridge():
    """Setup a network bridge for a machine with two nics."""
    interfaces_file = '/etc/network/interfaces'
    if not contains(interfaces_file, 'eth0 inet static'):
        sed(interfaces_file, 'iface eth0 inet dhcp',
            '#iface eth0 inet dhcp', use_sudo=True)
        append(interfaces_file, 'auto eth0', use_sudo=True)
        append(interfaces_file, 'iface eth0 inet static', use_sudo=True)
        append(interfaces_file, 'address 192.168.2.1', use_sudo=True)
        append(interfaces_file, 'netmask 255.255.255.0', use_sudo=True)
        append(interfaces_file, 'network 192.168.2.0', use_sudo=True)
        append(interfaces_file, 'broadcast 192.168.2.255', use_sudo=True)
        append(interfaces_file, 'gateway 192.168.2.1', use_sudo=True)
        sudo('ifdown eth0')
        sudo('ifup eth0')

    fabtools.require.deb.package('dhcp3-server')


@task
def masquerade():
    """Set up masquerading so that a box with two NICs can act as a router.
    """
    # Set up masquerade so that one host can act as a NAT gateway for a network
    # of hosts.
    #http://forums.fedoraforum.org/archive/index.php/t-178224.html
    rc_file = '/etc/rc.local'
    rule_1 = '/sbin/iptables -t nat -A POSTROUTING -o eth0 -j MASQUERADE'
    rule_2 = 'echo 1 > /proc/sys/net/ipv4/ip_forward'
    if not contains(rc_file, rule_1):
        sed(rc_file, '^exit 0', '#exit 0', use_sudo=True)
        append(rc_file, rule_1, use_sudo=True)
        append(rc_file, rule_2, use_sudo=True)



# coding=utf-8
"""Tools for deployment of cartodb on a server.

Note this is currently just a place holder. I plan to port these scripts to
fabric scripts:

https://github.com/nighthacker/cartodb-dev

Tim Sutton, August 2013"""



# coding=utf-8
"""Common helpers to bootstrap fabric."""
import os
from fabric.api import env, task, fastprint, run, hide
from fabric.utils import _AttributeDict as fdict
import fabtools


#TODO Not really used, but potentially useful
env.roledefs = {
    'test': ['localhost'],
    'dev': ['none@none.com'],
    'staging': ['none@none.com'],
    'production': ['none@none.com']
}

# Use ssh config if present
#if os.path.exists('~/.ssh/config'):
env.use_ssh_config = True
env.fg = None


def show_environment():
    """For diagnostics - show any pertinent info about server."""
    setup_env()
    fastprint('\n-------------------------------------------------\n')
    for key, value in env.fg.iteritems():
        fastprint('Key: %s \t\t Value: %s\n' % (key, value))
    fastprint('-------------------------------------------------\n')


def setup_env():
    """Things to do regardless of whether command is local or remote."""
    if env.fg is not None:
        fastprint('Environment already set!\n')
        return

    fastprint('Setting environment!\n')
    env.fg = fdict()
    with hide('output'):
        env.fg.user = run('whoami')
        # Workaround for
        env.fg.hostname = run('hostname')
        # this which fails in docker - see
        # https://github.com/dotcloud/docker/issues/1301
        #env.fg.hostname = fabtools.system.get_hostname()
        env.fg.home = os.path.join('/home/', env.fg.user)
        env.fg.workspace = os.path.join(env.fg.home, 'dev')
        env.fg.inasafe_git_url = 'git://github.com/AIFDR/inasafe.git'
        env.fg.qgis_git_url = 'git://github.com/qgis/QGIS.git'
        env.fg.kandan_git_url = 'git://github.com/kandanapp/kandan.git'
        env.fg.gdal_svn_url = 'https://svn.osgeo.org/gdal/trunk/gdal'
        env.fg.tilemill_tarball_url = (
            'http://tilemill.s3.amazonaws.com/latest/install-tilemill.tar.gz')
        env.fg.inasafe_checkout_alias = 'inasafe-fabric'
        env.fg.qgis_checkout_alias = 'qgis-fabric'
        env.fg.inasafe_code_path = os.path.join(
            env.fg.workspace, env.fg.inasafe_checkout_alias)
        env.fg.qgis_code_path = os.path.join(
            env.fg.workspace, env.fg.qgis_checkout_alias)


@task
def add_ubuntugis_ppa():
    """Ensure we have ubuntu-gis repo."""
    fabtools.deb.update_index(quiet=True)
    fabtools.require.deb.package('software-properties-common')
    fabtools.require.deb.ppa(
        #'ppa:ubuntugis/ubuntugis-unstable', auto_yes=True)
        'ppa:ubuntugis/ubuntugis-unstable')



# coding=utf-8
"""Setup file for distutils / pypi."""
from distutils.core import setup

setup(
    name='fabgis',
    version='0.16.0',
    author='Tim Sutton',
    author_email='tim@linfiniti.com',
    packages=['fabgis', ],
    data_files=[],
    scripts=[],
    url='http://pypi.python.org/pypi/fabgis/',
    license='LICENSE.txt',
    description='Fabric tools for a busy FOSSGIS user.',
    long_description=open('README.txt').read(),
    install_requires=[
        "Fabric == 1.6.0",
        "fabtools >= 0.15.0"
    ],
)



# coding=utf-8
"""Tasks for setting up sphinx."""
import fabtools
from fabric.api import task, sudo, fastprint
from fabric.colors import green, blue
from fabtools.deb import is_installed


@task
def setup_latex():
    """Install latex and friends needed to generate sphinx PDFs."""
    fastprint(blue('Setting up LaTeX\n'))
    fabtools.deb.update_index(quiet=True)
    fabtools.require.deb.package('texlive-latex-extra')
    fabtools.require.deb.package('texinfo')
    fabtools.require.deb.package('texlive-fonts-recommended')
    fastprint(green('Setting up LaTeX completed\n'))


@task
def setup_sphinx():
    """Install sphinx from pip.

    We prefer packages from pip as ubuntu packages are usually old.
    To build the Documentation we also need to check and update the
    subjacent docutils installation"""
    fastprint(blue('Setting up Sphinx\n'))
    if is_installed('docutils-common'):
        sudo('apt-get remove docutils-common')
    if is_installed('docutils-doc'):
        sudo('apt-get remove docutils-doc')
    if is_installed('python-docutils'):
        sudo('apt-get remove python-docutils')
    sudo('pip install --upgrade docutils==0.10')
    sudo('pip install sphinx')
    fastprint(green('Setting up Sphinx completed\n'))


@task
def setup_transifex():
    """Install transifex client."""
    fastprint(blue('Setting up transifex command line client\n'))
    sudo('pip install transifex-client')
    fastprint(green('Setting up transifex command line client completed\n'))



# coding=utf-8
"""Tools relating tot he use of python virtualenv."""

import os
from fabric.api import fastprint, run, cd, task
from fabric.contrib.files import sed
from fabric.colors import blue, green
from fabtools.require.deb import packages as require_packages
from fabtools.require.python import virtualenv

from .common import setup_env


@task
def setup_venv(code_path, requirements_file='requirements.txt'):
    """Initialise or update the virtual environment.

    It will also ensure build-essential is installed, though you may need to
    install required dev library using your own script.

    :param code_path: Base directory under which the venv dir should be made.
    :type code_path: str

    :param requirements_file: Name of the requirements file to use in the
        base directory. Defaults to ``requirements.txt``.
    :type requirements_file: str

    To run e.g.::

        fab -H 192.168.1.1:2222 setup_venv

    """
    setup_env()
    fastprint(blue('Setting up virtual env in: \n%s\nusing\n%s' % (
        code_path, requirements_file)))
    require_packages(['python-virtualenv', 'build-essential'])
    with cd(code_path):
        # Ensure we have a venv set up
        virtualenv('venv')
        run('venv/bin/pip install -r %s' % requirements_file)
    fastprint(green('Virtualenv setup completed.'))


@task
def build_pil(code_path):
    """Build pil with proper support for jpeg etc.

    :param code_path: Directory where the code lives.
    :type code_path: str

    .. note:: Any existing PIL will be uninstalled.

    .. versionadded: 0.16.0
    """
    require_packages(['libjpeg-dev', 'libfreetype6', 'libfreetype6-dev'])

    tcl = 'TCL_ROOT = None'
    jpg = 'JPEG_ROOT = None'
    zlib = 'ZLIB_ROOT = None'
    tiff = 'TIFF_ROOT = None'
    freetype = 'FREETYPE_ROOT = None'

    tcl_value = (
        'TCL_ROOT = "/usr/lib/x86_64-linux-gnu/", "/usr/include"')
    jpg_value = (
        'JPEG_ROOT = "/usr/lib/x86_64-linux-gnu/", "/usr/include"')
    zlib_value = (
        'ZLIB_ROOT = "/usr/lib/x86_64-linux-gnu/", "/usr/include"')
    tiff_value = (
        'TIFF_ROOT = "/usr/lib/x86_64-linux-gnu/", "/usr/include"')
    freetype_value = (
        'FREETYPE_ROOT = "/usr/lib/x86_64-linux-gnu/", "/usr/include"')

    venv = os.path.join(code_path, 'venv')
    with cd(venv):
        run('bin/pip uninstall pil')
        run('wget -c http://effbot.org/downloads/Imaging-1.1.7.tar.gz')
        run('tar xfz Imaging-1.1.7.tar.gz')
        with cd(os.path.join(venv, 'Imaging-1.1.7')):
            sed('setup.py', tcl, tcl_value)
            sed('setup.py', jpg, jpg_value)
            sed('setup.py', zlib, zlib_value)
            sed('setup.py', tiff, tiff_value)
            sed('setup.py', freetype, freetype_value)
            run('../bin/python setup.py install')


@task
def build_python_gdal(code_path):
    """Build python gdal in a virtualenv etc.

    :param code_path: Directory where the code lives.
    :type code_path: str

    .. note:: Any existing GDAL will be uninstalled.

    .. versionadded: 0.16.0
    """

    # Gdal does not build cleanly from requirements so we follow advice
    # from http://ubuntuforums.org/showthread.php?t=1769445
    pip_path = os.path.join(code_path, 'venv', 'bin', 'pip')
    gdal_build_path = os.path.join(code_path, 'venv', 'build', 'GDAL')

    result = run('%s install --no-install GDAL' % pip_path)
    if 'Requirement already satisfied ' not in result:
        with cd(gdal_build_path):
            run('python setup.py build_ext --include-dirs=/usr/include/gdal')
            run('%s install --no-download GDAL' % pip_path)



# coding=utf-8
"""Helpers for installing hdf5"""

import fabtools
from fabric.contrib.files import exists
from fabric.api import fastprint, run, cd, env, task, sudo, settings

from .system import setup_ccache
from .common import setup_env
from .utilities import append_if_not_present


@task
def build_hdf5(version='1.8.11'):
    """Get proj4 from tarball and build it.

    :param version: hdf5 version to build. The version should be consistent
        with a downloadable tar file from the project web site. Default is
        the current stable release.
    :type version: str
    """
    setup_env()
    fabtools.require.deb.package('build-essential')
    setup_ccache()

    code_base = '%s/cpp' % env.fg.workspace
    filename = 'hdf5-%s' % version
    source_url = (
        'http://www.hdfgroup.org/ftp/HDF5/current/src/%s.tar.gz' %
        filename)
    code_path = '%s/%s' % (code_base, filename)

    if not exists(code_path):
        fastprint('Extracted tarball does not exist, creating.')
        with cd(code_base):
            if exists('%s.tar.gz' % filename):
                run('rm %s.tar.gz' % filename)
            run('wget %s' % source_url)
            run('tar xfz %s.tar.gz' % filename)

    processor_count = run('cat /proc/cpuinfo | grep processor | wc -l')

    with cd(code_path):
        # Dont fail if make clean does not work
        with settings(warn_only=True):
            run('make clean')
        run('./configure')
        run('make -j %s' % processor_count)
        sudo('make install')
    # Write to ld path too so libs are loaded nicely
    ld_file = '/etc/ld.so.conf.d/usr_local_lib.conf'
    with settings(warn_only=True):
        sudo('rm %s' % ld_file)
    append_if_not_present(ld_file, '/usr/local/lib', use_sudo=True)
    sudo('ldconfig')
